from pulp import *

class Reservation(object):
    """
    class representing a reservation in the system
    """
    def __init__(self, name, datarate, priority):
        self.name = name
        self.datarate = datarate
        self.priority = priority
        self.lpvar = LpVariable(name, cat='Binary')
        self.lpvar.originalReservation = self # save a back reference for later

    @property
    def online(self):
        """
        return whether this path is to be considered online
        """
        return self.lpvar.varValue == 1
        
    def __repr__(self):
        return '%s(%d, %d)'%(self.name, self.datarate, self.priority)
        
class Relation(object):
    def __init__(self, reservations, typename='Relation'):
        if len(reservations) < 2:
            raise Exception('Relation has to contain at least two paths')
        self.reservations = reservations
        self.lpconstraints = []
        self.typename = typename
    
    def __str__(self):
        return '%s(%s)'%(self.typename, ', '.join((resv.name for resv in self.reservations)))
    
    def __repr__(self):
        return str(self)
        
        
class Exclusive(Relation):
    def __init__(self, reservations):
        super().__init__(reservations, 'Exclusive')
        cond = self.reservations[0].lpvar
        # create the rest of the relation
        for resv in self.reservations[1:]:
            cond += resv.lpvar
        cond = cond <= 1
        self.lpconstraints.append(cond)
        
    @property
    def valid(self):
        return len([res for res in self.reservations if res.online]) <= 1 # the relation is valid, if at most one is online
    
class AtLeastN(Relation):
    def __init__(self, reservations, num):
        super().__init__(reservations, 'AtLeastN')
        self.number = num
        self._helpervar = LpVariable('h-%s-%s'%(self.reservations[0].name,self.reservations[-1].name)) # a helper variable counting the number of enabled reservations
        cond = self.reservations[0].lpvar
        self.lpconstraints.append(self._helpervar >= self.reservations[0].lpvar * self.number)
        for resv in self.reservations[1:]:
            cond += resv.lpvar
            self.lpconstraints.append(self._helpervar >= resv.lpvar * self.number) # the h >= resv * Num condition
        cond = cond == self._helpervar # the big counting condition over all enabled reservations
        self.lpconstraints.append(cond)

    @property
    def valid(self):
        online = len([res for res in self.reservations if res.online])
        return online == 0 or online >= self.number
    
class MutualDependency(AtLeastN):
    def __init__(self, reservations):
        super().__init__(reservations, len(reservations))
        self.typename = 'Mutual'

class Dependent(Relation):
    """
    Dependency relation where all but the first elements depend on the first
    """
    def __init__(self, reservations):
        super().__init__(reservations, 'Dependent')
        for resv in self.reservations[1:]:
            self.lpconstraints.append(self.reservations[0].lpvar >= resv.lpvar)
            
    @property
    def valid(self):
        return not any((resv.online for resv in self.reservations[1:])) or self.reservations[0].online
    
def optimize(reservations, relations, target_capacity):
    """
    do a single optimization run
    """
    # setup the standard knapsack problem
    prob = LpProblem('WastedCapacity', LpMaximize)
    profit = reservations[0].lpvar*reservations[0].priority
    cost = reservations[0].lpvar*reservations[0].datarate
    for resv in reservations[1:]:
        profit += resv.lpvar*resv.priority
        cost   += resv.lpvar*resv.datarate
    prob += profit
    prob += cost <= target_capacity
    # add relation constraints, if any
    for rel in relations:
        for constraint in rel.lpconstraints:
            prob += constraint
    # run the solver
    if COIN().solve(prob) < 0:
        with open('solver-errors', 'a') as errorfile:
            print('########################################', file=errorfile)
            print(reservations, file=errorfile)
            print(relations, file=errorfile)
            print(target_capacity, file=errorfile)
            print(prob, file=errorfile)
            print('########################################', file=errorfile)
        raise Exception("Couldn't find a valid solution. This should not happen in the simulation")
    # return the original reservations, which were enabled
    return [v.originalReservation for v in prob.variables() if hasattr(v, 'originalReservation') and v.varValue==1]