from pulp import *
from random import randint
import timeit
import pandas
import multiprocessing as mp
import numpy
import kasymosa

def analyzeResult(online_reservations, relations, target_capacity):
    """
    analyze the optimization result for the statistics like wasted capacity etc.
    """
    def _containing(reservation, relations):
        """
        return the containing relations, if any
        """
        return [rel for rel in relations if (reservation in rel.reservations)]
        
    def _isWasted(reservation, relations):
        """
        check, whether a reservation is wasted, i.e. it is part of a relation, 
        that is offline (and it itself is online)
        """
        containing_relation = _containing(reservation, relations)
        if len(containing_relation) == 0:
            print('Have single path')
            return False # reservations without relation are not wasted by definition
        else:
            return not containing_relation[0].valid # if the containing relation is offline, we're screwed
    
    def _isOnline(session, relations):
        """
        check whether a path/relation is really online. Standalone online paths are 
        obviously online. Relations must not only be valid, but also have resources 
        assigned (i.e. at least one member reservation is online)
        """
        if type(session) is kasymosa.Reservation: # reservations, that are not contained in a relation are counted as online
            return len(_containing(session, relations))==0 and session.online
        else:
            return session.valid and any((res.online for res in session.reservations)) # relations are online if they are valid and at least one reservation is online

    def _couldBeOnline(session, relations):
        """
        check whether the session could be used at all, i.e. it has at least one path online
        This is only meaningful in the context of the Exclusive relation where we could argue, that
        even with multiple online paths the session still can be used (although anything but the best path would be wasted).
        """
        if type(session) is kasymosa.Reservation:
            return len(_containing(session, relations)) == 0 and session.online
        else:
            return any((res.online for res in session.reservations))
        
    return {
        "online_reservations" : len(online_reservations),
        "used_capacity" : sum((resv.datarate for resv in online_reservations)),
        "wasted_capacity" : sum((resv.datarate for resv in online_reservations if _isWasted(resv,relations))),
        "online_sessions" : len([res for res in online_reservations if _isOnline(res, relations)]) + len([rel for rel in relations if _isOnline(rel, None)]), # online sessions are single paths or relations
        "could_be_sessions" : len([res for res in online_reservations if _couldBeOnline(res, relations)]) + len([rel for rel in relations if _couldBeOnline(rel, None)])
    }
                        
def runSimulation(num_reservations, relation_size, remaining_capacity_ratio, relation_type, value_variance):
    """
    run a single simulation instance
    """
    # create the necessary amount of reservations
    reservations = [kasymosa.Reservation('res%d'%i, randint(1000, 100000), randint(1, value_variance)) for i in range(num_reservations)]
    # create reservations as needed
    relations = [relation_type(reservations[i:i+relation_size]) for i in range(0, num_reservations, relation_size)]
    # create the time correlated relations (all relations are assumed to be mutual dependencies)
    ct_relations = [kasymosa.MutualDependency(reservations[i:i+relation_size]) for i in range(0, num_reservations, relation_size)]
    # calculate the necessary capacities
    capacity_before = sum((resv.datarate for resv in reservations))
    target_capacity = capacity_before * remaining_capacity_ratio
    # do the optimization including the reservations
    kasymosa_result = kasymosa.optimize(reservations, relations, target_capacity)
    kasymosa_statistics = analyzeResult(kasymosa_result, relations, target_capacity)
    # do it again excluding the reservations
    standard_result = kasymosa.optimize(reservations, [], target_capacity)
    standard_statistics = analyzeResult(standard_result, relations, target_capacity)
    # and one last time based on time-correlation
    correlated_result = kasymosa.optimize(reservations, ct_relations, target_capacity)
    correlated_statistics = analyzeResult(correlated_result, relations, target_capacity) # we optimize with the ct-relations, but analyze with the "normal" ones to get the discrepancies for that model
    kasymosa_statistics['initial_capacity'] = capacity_before
    standard_statistics['initial_capacity'] = capacity_before
    correlated_statistics['initial_capacity'] = capacity_before
    return {
        "kasymosa" : kasymosa_statistics,
        "standard" : standard_statistics,
        "correlated" : correlated_statistics
    }

def relCreator(cls, *args):
    def _con(relations):
        return cls(relations, *args)
    return _con

def proc_main(param):
    """
    main function for the subprocess running the simulation for a specific number of paths
    """
    num_reservations, relation_size = param
    results = []
    for remaining_capacity_ratio in [0.9, 0.8, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1]:
        for relation_type in [('Mutual', relCreator(kasymosa.MutualDependency)), 
                                ('Exclusive', relCreator(kasymosa.Exclusive)), 
                                ('AtLeastN', relCreator(kasymosa.AtLeastN, int(relation_size/2))),
                                ('Dependent', relCreator(kasymosa.Dependent))]:
            for value_variance in [1, 2, 8, 32, 128, 255]:
                if num_reservations >= relation_size:
                    print(('running: %d reservations, %d relation size,'+
                            ' %d%% remaining capacity, %s, %d value variance')%(num_reservations, \
                                                                                relation_size,\
                            remaining_capacity_ratio*100, relation_type[0], value_variance))
                    repetition = 0
                    while repetition < 100:
                        try:
                            result = runSimulation(num_reservations, relation_size, remaining_capacity_ratio, relation_type[1], value_variance)
                            results.append([
                                num_reservations,
                                relation_size,
                                remaining_capacity_ratio,
                                relation_type[0],
                                value_variance,
                                'kasymosa',
                                repetition,
                                result['kasymosa']['initial_capacity'],
                                result['kasymosa']['online_reservations'],
                                result['kasymosa']['used_capacity'],
                                result['kasymosa']['wasted_capacity'],
                                result['kasymosa']['online_sessions'],
                                result['kasymosa']['could_be_sessions']
                            ])
                            results.append([
                                num_reservations,
                                relation_size,
                                remaining_capacity_ratio,
                                relation_type[0],
                                value_variance,
                                'standard',
                                repetition,
                                result['standard']['initial_capacity'],
                                result['standard']['online_reservations'],
                                result['standard']['used_capacity'],
                                result['standard']['wasted_capacity'],
                                result['standard']['online_sessions'],
                                result['standard']['could_be_sessions']
                            ])
                            results.append([
                                num_reservations,
                                relation_size,
                                remaining_capacity_ratio,
                                relation_type[0],
                                value_variance,
                                'correlated',
                                repetition,
                                result['correlated']['initial_capacity'],
                                result['correlated']['online_reservations'],
                                result['correlated']['used_capacity'],
                                result['correlated']['wasted_capacity'],
                                result['correlated']['online_sessions'],
                                result['correlated']['could_be_sessions']
                            ])
                            repetition += 1
                        except PulpSolverError:
                            print('Stumbled upon solver error. Retrying.')
    result_frame = pandas.DataFrame(data=results, columns=['#reservations', 'relation size', \
                                                            'remaining capacity ratio', 'relation type', \
                                                            'value variance', 'system type', 'run', \
                                                            'initial capacity', 'remaining reservations', 'used capacity',\
                                                            'wasted capacity', 'online sessions', 'could-be sessions'])                                
    result_frame.to_csv('wasted-capacity-offline-sessions-%d-%d.csv'%(num_reservations, relation_size))

if __name__=='__main__':
    pool = mp.Pool(processes=4)
    # The simulation is rather big, running a total of ~560000 parameter combinations
    params = []
    for num_reservations in [8, 16, 32, 128, 512]:
        for relation_size in [2,4,8,16,32]:
            if num_reservations >= relation_size:
                params.append((num_reservations, relation_size))
    pool.map(proc_main, params)